class TributeContext {
    constructor(tribute) {
        this.tribute = tribute;
        this.mentionText = '';
        this.externalTrigger = false;
    }
    process(scrollTo) {
        if (this.tribute.menu.element === null || !this.collection)
            return;
        const ul = this.tribute.menu.element.querySelector('ul');
        if (ul === null)
            throw new Error('menu do not have "ul" element');
        const collection = this.collection;
        const processor = (values) => {
            // Tribute may not be active any more by the time the value callback returns
            if (!this.tribute.isActive) {
                return;
            }
            const items = this._filterItems(collection, values);
            this.filteredItems = items;
            if (!items.length) {
                this._handleNoItem(ul, collection, scrollTo);
            }
            else {
                this._renderMenu(items, ul, collection, scrollTo);
            }
        };
        if (typeof collection.values === 'function') {
            if (collection.loadingItemTemplate) {
                ul.innerHTML = collection.loadingItemTemplate;
                this.tribute.range.positionMenuAtCaret(scrollTo);
            }
            collection.values(this.mentionText, processor);
        }
        else if (collection.values !== null) {
            processor(collection.values);
        }
    }
    updateSelection(info) {
        this.selectedPath = info.mentionSelectedPath;
        this.mentionText = info.mentionText || '';
        this.selectedOffset = info.mentionSelectedOffset;
    }
    get isMentionLengthUnderMinimum() {
        if (!this.collection)
            return undefined;
        return this.mentionText.length < this.collection.menuShowMinLength;
    }
    _filterItems(collection, values) {
        const opts = collection.searchOpts;
        const lookup = collection.lookup;
        const items = this.tribute.search.filter(this.mentionText, values, {
            pre: opts.pre || '<span>',
            post: opts.post || '</span>',
            skip: opts.skip || false,
            caseSensitive: opts.caseSensitive || false,
            extract: (el) => {
                if (typeof lookup === 'string') {
                    return el[lookup];
                }
                if (typeof lookup === 'function') {
                    return lookup(el, this.mentionText);
                }
                throw new Error('Invalid lookup attribute, lookup must be string or function.');
            },
        });
        if (collection.menuItemLimit) {
            return items.slice(0, collection.menuItemLimit);
        }
        return items;
    }
    _handleNoItem(ul, collection, scrollTo) {
        if (!this.element)
            throw new Error('element is empty');
        const noMatchEvent = new CustomEvent('tribute-no-match', {
            detail: this.tribute.menu.element,
        });
        this.element.dispatchEvent(noMatchEvent);
        if ((typeof collection.noMatchTemplate === 'function' && !collection.noMatchTemplate()) || !collection.noMatchTemplate) {
            this.tribute.hideMenu();
        }
        else {
            ul.innerHTML = typeof collection.noMatchTemplate === 'function' ? collection.noMatchTemplate() : collection.noMatchTemplate;
            this.tribute.range.positionMenuAtCaret(scrollTo);
        }
    }
    _renderMenu(items, ul, collection, scrollTo) {
        ul.innerHTML = '';
        const doc = this.tribute.range.getDocument();
        const fragment = doc.createDocumentFragment();
        this.tribute.menu.selected = items.findIndex((item) => item.original.disabled !== true);
        items.forEach((item, index) => {
            const li = this._createMenuItem(item, collection, index, doc);
            fragment.appendChild(li);
        });
        ul.appendChild(fragment);
        this.tribute.range.positionMenuAtCaret(scrollTo);
    }
    _createMenuItem(item, collection, index, doc) {
        const li = doc.createElement('li');
        li.setAttribute('data-index', index.toString());
        if (item.original.disabled) {
            li.setAttribute('data-disabled', 'true');
        }
        li.className = collection.itemClass;
        li.addEventListener('mousemove', (e) => {
            const [li, index] = this._findLiTarget(e.target);
            if ('movementY' in e && e.movementY !== 0 && index !== null && typeof index !== 'undefined') {
                this.tribute.menu.setActiveLi(Number.parseInt(index));
            }
        });
        if (this.tribute.menu.selected === index) {
            li.classList.add(collection.selectClass);
        }
        // remove all content in the li and append the content of menuItemTemplate
        const menuItemDomOrString = collection.menuItemTemplate(item);
        if (menuItemDomOrString instanceof Element) {
            li.innerHTML = '';
            li.appendChild(menuItemDomOrString);
        }
        else {
            li.innerHTML = menuItemDomOrString;
        }
        return li;
    }
    _findLiTarget(el) {
        if (!el || !(el instanceof HTMLElement))
            return [];
        const index = el.getAttribute('data-index');
        return !index ? this._findLiTarget(el.parentNode) : [el, index];
    }
}

function addHandler(element, ...args) {
    element.addEventListener(...args);
    return () => element.removeEventListener(...args);
}
function isContentEditable(element) {
    return element.nodeName !== 'INPUT' && element.nodeName !== 'TEXTAREA';
}
function isNotContentEditable(element) {
    return element instanceof HTMLInputElement || element instanceof HTMLTextAreaElement;
}

const hotkeys = ['tab', 'backspace', 'enter', 'escape', 'space', 'arrowup', 'arrowdown'];
class TributeEvents {
    constructor(tribute) {
        this.tribute = tribute;
        this.removers = [];
        this.inputEvent = false;
    }
    bind(element) {
        this.removers.push(addHandler(element, 'keydown', (event) => {
            this.keydown(event);
        }));
        this.removers.push(addHandler(element, 'keyup', (event) => {
            this.keyup(event);
        }));
        this.removers.push(addHandler(element, 'input', (event) => {
            this.input(event);
        }));
    }
    unbind(element) {
        for (const remover of this.removers) {
            remover();
        }
    }
    keydown(event) {
        if (!(event instanceof KeyboardEvent))
            return;
        const element = event.currentTarget;
        if (this.shouldDeactivate(event)) {
            this.tribute.isActive = false;
            this.tribute.hideMenu();
        }
        this.commandEvent = false;
        const key = getCode(event.key);
        if (isHotkey(key) && element instanceof HTMLElement) {
            this.commandEvent = true;
            this.callbacks[key](event, element);
        }
    }
    input(event) {
        event.currentTarget;
        this.inputEvent = true;
        this.keyup(event);
    }
    keyup(event) {
        if (!(event instanceof KeyboardEvent))
            return;
        const element = event.currentTarget;
        if (!(element instanceof HTMLElement))
            return;
        if (this.inputEvent) {
            this.inputEvent = false;
        }
        this.updateSelection(element);
        if (!event.key || event.key === 'Escape')
            return;
        if (!this.tribute.allowSpaces && this.tribute.hasTrailingSpace) {
            this.tribute.hasTrailingSpace = false;
            this.commandEvent = true;
            this.callbacks.space(event, element);
            return;
        }
        if (!this.tribute.isActive) {
            if (this.tribute.autocompleteMode) {
                this.triggerChar(event, element, '');
            }
            else {
                const charCode = this.getTriggerCharCode();
                if (Number.isNaN(charCode) || !charCode)
                    return;
                const trigger = this.tribute.triggers().find((trigger) => {
                    return (trigger === null || trigger === void 0 ? void 0 : trigger.charCodeAt(0)) === charCode;
                });
                if (typeof trigger !== 'undefined') {
                    this.triggerChar(event, element, trigger);
                }
            }
        }
        if (this.tribute.current.isMentionLengthUnderMinimum) {
            this.tribute.hideMenu();
            return;
        }
        if (((this.tribute.current.trigger || this.tribute.autocompleteMode) && this.commandEvent === false) || this.showMenuOnBackspace(event.key)) {
            this.tribute.showMenuFor(element, true);
        }
    }
    shouldDeactivate(event) {
        var _a;
        if (!this.tribute.isActive)
            return false;
        if (!(event instanceof KeyboardEvent))
            return false;
        if (((_a = this.tribute.current.mentionText) === null || _a === void 0 ? void 0 : _a.length) === 0) {
            let eventKeyPressed = false;
            const key = getCode(event.key);
            if (isHotkey(key)) {
                eventKeyPressed = true;
            }
            return !eventKeyPressed;
        }
        return false;
    }
    getTriggerCharCode() {
        const tribute = this.tribute;
        const info = tribute.range.getTriggerInfo(false, tribute.hasTrailingSpace, true, tribute.allowSpaces, tribute.autocompleteMode);
        if (info === null || info === void 0 ? void 0 : info.mentionTriggerChar) {
            return info.mentionTriggerChar.charCodeAt(0);
        }
        return false;
    }
    updateSelection(el) {
        this.tribute.current.element = el;
        const info = this.tribute.range.getTriggerInfo(false, this.tribute.hasTrailingSpace, true, this.tribute.allowSpaces, this.tribute.autocompleteMode);
        if (info) {
            this.tribute.current.updateSelection(info);
        }
    }
    triggerChar(e, el, trigger) {
        const tribute = this.tribute;
        tribute.current.trigger = trigger;
        const collectionItem = tribute.collection.find((item) => {
            return item.trigger === trigger;
        });
        tribute.current.collection = collectionItem;
        if (!tribute.current.isMentionLengthUnderMinimum && this.inputEvent) {
            tribute.showMenuFor(el, true);
        }
    }
    get callbacks() {
        if (!this._callbacks) {
            this._callbacks = {
                enter: (e, el) => {
                    var _a;
                    // choose selection
                    const filteredItems = this.tribute.current.filteredItems;
                    if (this.tribute.isActive && filteredItems && filteredItems.length) {
                        e.preventDefault();
                        e.stopPropagation();
                        if (((_a = this.tribute.current.filteredItems) === null || _a === void 0 ? void 0 : _a.length) === 0) {
                            this.tribute.menu.unselect();
                        }
                        setTimeout(() => {
                            this.tribute.selectItemAtIndex(this.tribute.menu.selected.toString(), e);
                            this.tribute.hideMenu();
                        }, 0);
                    }
                },
                escape: (e, el) => {
                    if (this.tribute.isActive) {
                        e.preventDefault();
                        e.stopPropagation();
                        this.tribute.isActive = false;
                        this.tribute.hideMenu();
                    }
                },
                tab: (e, el) => {
                    // choose first match
                    this.callbacks.enter(e, el);
                },
                space: (e, el) => {
                    if (this.tribute.isActive) {
                        if (this.tribute.spaceSelectsMatch) {
                            this.callbacks.enter(e, el);
                        }
                        else if (!this.tribute.allowSpaces) {
                            e.stopPropagation();
                            setTimeout(() => {
                                this.tribute.hideMenu();
                                this.tribute.isActive = false;
                            }, 0);
                        }
                    }
                },
                arrowup: (e, el) => {
                    // navigate up ul
                    if (this.tribute.isActive && this.tribute.current.filteredItems) {
                        e.preventDefault();
                        e.stopPropagation();
                        const count = this.tribute.current.filteredItems.length;
                        this.tribute.menu.up(count);
                    }
                },
                arrowdown: (e, el) => {
                    // navigate down ul
                    if (this.tribute.isActive && this.tribute.current.filteredItems) {
                        e.preventDefault();
                        e.stopPropagation();
                        const count = this.tribute.current.filteredItems.length;
                        this.tribute.menu.down(count);
                    }
                },
                backspace: (e, el) => {
                    if (this.tribute.isActive) {
                        if (this.tribute.current && this.tribute.current.mentionText.length < 1) {
                            this.tribute.hideMenu();
                        }
                        else {
                            this.tribute.showMenuFor(el);
                        }
                    }
                },
            };
        }
        return this._callbacks;
    }
    showMenuOnBackspace(key) {
        const isBackspace = key === 'Backspace';
        return isNotContentEditable(this.tribute.current.element) ? isBackspace : this.tribute.isActive && isBackspace;
    }
}
function getCode(key) {
    return key === ' ' ? 'space' : key.toLowerCase();
}
function includes(array, input) {
    return array.includes(input);
}
function isHotkey(code) {
    return includes(hotkeys, code);
}

class TributeMenu {
    constructor(tribute) {
        this.tribute = tribute;
        this.element = null;
        this.selected = 0;
    }
    create(doc, containerClass) {
        const wrapper = doc.createElement('div');
        const ul = doc.createElement('ul');
        wrapper.className = containerClass;
        wrapper.appendChild(ul);
        this.element = this.tribute.menuContainer ? this.tribute.menuContainer.appendChild(wrapper) : doc.body.appendChild(wrapper);
        return this.element;
    }
    activate() {
        this.selected = 0;
        window.setTimeout(() => {
            if (this.element === null)
                throw new Error('the menu element is null!');
            this.element.scrollTop = 0;
        }, 0);
    }
    deactivate() {
        if (this.element === null)
            throw new Error('the menu element is null!');
        this.element.style.cssText = 'display: none;';
        this.selected = 0;
    }
    unselect() {
        this.selected = -1;
    }
    get isActive() {
        return this.element !== null;
    }
    get items() {
        if (this.element === null)
            throw new Error('the menu element is null!');
        return this.element.querySelectorAll('li');
    }
    get itemDisabled() {
        var _a;
        return ((_a = this.items[this.selected]) === null || _a === void 0 ? void 0 : _a.getAttribute('data-disabled')) === 'true';
    }
    up(count) {
        if (this.element === null)
            throw new Error('the menu element is null!');
        //If menu.selected is -1 then there are no valid, non-disabled items
        //to navigate through
        if (this.selected === -1) {
            return;
        }
        do {
            this.selected--;
            if (this.selected === -1) {
                this.selected = count - 1;
                this.element.scrollTop = this.element.scrollHeight;
            }
        } while (this.itemDisabled);
        this.setActiveLi();
    }
    down(count) {
        if (this.element === null)
            throw new Error('the menu element is null!');
        //If menu.selected is -1 then there are no valid, non-disabled items
        //to navigate through
        if (this.selected === -1) {
            return;
        }
        do {
            this.selected++;
            if (this.selected >= count) {
                this.selected = 0;
                this.element.scrollTop = 0;
            }
        } while (this.itemDisabled);
        this.setActiveLi();
    }
    setActiveLi(index) {
        if (this.element === null)
            throw new Error('the menu element is null!');
        if (!this.tribute.current.collection)
            return;
        const selectClass = this.tribute.current.collection.selectClass;
        const lis = this.items;
        lis.length >>> 0;
        if (index) {
            this.selected = index;
        }
        const element = this.element;
        this.items.forEach((li, i) => {
            if (i === this.selected) {
                if (li.getAttribute('data-disabled') !== 'true') {
                    li.classList.add(selectClass);
                }
                const liClientRect = li.getBoundingClientRect();
                const menuClientRect = element.getBoundingClientRect();
                if (liClientRect.bottom > menuClientRect.bottom) {
                    const scrollDistance = liClientRect.bottom - menuClientRect.bottom;
                    element.scrollTop += scrollDistance;
                }
                else if (liClientRect.top < menuClientRect.top) {
                    const scrollDistance = menuClientRect.top - liClientRect.top;
                    element.scrollTop -= scrollDistance;
                }
            }
            else {
                li.classList.remove(selectClass);
            }
        });
    }
    positionAtCaret(info, coordinates) {
        if (this.element === null)
            throw new Error('the menu element is null!');
        if (typeof info === 'undefined') {
            this.element.style.cssText = 'display: none';
            return;
        }
        if (!this.tribute.positionMenu) {
            this.element.style.cssText = 'display: block;';
            return;
        }
        this.element.style.cssText = `top: ${coordinates.top}px;
                                   left: ${coordinates.left}px;
                                   right: ${coordinates.right}px;
                                   bottom: ${coordinates.bottom}px;
                                   max-height: ${coordinates.maxHeight || 500}px;
                                   max-width: ${coordinates.maxWidth || 300}px;
                                   position: ${coordinates.position || 'absolute'};
                                   display: block;`;
        if (coordinates.left === 'auto') {
            this.element.style.left = 'auto';
        }
        if (coordinates.top === 'auto') {
            this.element.style.top = 'auto';
        }
    }
    getDimensions() {
        if (this.element === null)
            throw new Error('the menu element is null!');
        // Width of the menu depends of its contents and position
        // We must check what its width would be without any obstruction
        // This way, we can achieve good positioning for flipping the menu
        const dimensions = {
            width: null,
            height: null,
        };
        this.element.style.cssText = `top: 0px;
                                 left: 0px;
                                 position: fixed;
                                 display: block;
                                 visibility; hidden;
                                 max-height:500px;`;
        dimensions.width = this.element.offsetWidth;
        dimensions.height = this.element.offsetHeight;
        this.element.style.cssText = 'display: none;';
        return dimensions;
    }
}

class TributeMenuEvents {
    constructor(tribute) {
        this.tribute = tribute;
        this.removers = [];
    }
    bind(menu) {
        const menuContainerScrollEvent = this.debounce(() => {
            if (this.tribute.isActive) {
                this.tribute.hideMenu();
            }
        }, 10, false);
        const windowResizeEvent = this.debounce(() => {
            if (this.tribute.isActive) {
                this.tribute.hideMenu();
            }
        }, 10, false);
        const closeOnScrollEvent = this.debounce(() => {
            if (this.tribute.isActive) {
                this.tribute.hideMenu();
            }
        }, 10, false);
        this.removers.push(addHandler(this.tribute.range.getDocument(), 'mousedown', (event) => this.click(event), false));
        this.removers.push(addHandler(window, 'resize', windowResizeEvent));
        if (this.tribute.closeOnScroll === true) {
            this.removers.push(addHandler(window, 'scroll', closeOnScrollEvent));
        }
        else if (this.tribute.closeOnScroll !== false) {
            this.removers.push(addHandler(this.tribute.closeOnScroll, 'scroll', closeOnScrollEvent, false));
        }
        else {
            if (this.tribute.menuContainer) {
                this.removers.push(addHandler(this.tribute.menuContainer, 'scroll', menuContainerScrollEvent, false));
            }
            else {
                this.removers.push(addHandler(window, 'scroll', menuContainerScrollEvent));
            }
        }
    }
    unbind(menu) {
        for (const remover of this.removers) {
            remover();
        }
    }
    click(event) {
        var _a, _b;
        const element = event.target;
        const tribute = this.tribute;
        if (!tribute.current || !(element instanceof Node))
            return;
        if ((_a = tribute.menu.element) === null || _a === void 0 ? void 0 : _a.contains(element)) {
            let li = element;
            event.preventDefault();
            event.stopPropagation();
            while (li.nodeName.toLowerCase() !== 'li') {
                li = li.parentNode;
                if (!li || li === tribute.menu.element) {
                    // When li === tribute.menu, it's either a click on the entire component or on the scrollbar (if visible)
                    li = undefined;
                    break;
                }
            }
            if (!(li instanceof HTMLElement))
                return;
            if (li.getAttribute('data-disabled') === 'true') {
                return;
            }
            if (((_b = tribute.current.filteredItems) === null || _b === void 0 ? void 0 : _b.length) === 0) {
                li.setAttribute('data-index', '-1');
            }
            const index = li.getAttribute('data-index');
            if (index !== null) {
                tribute.selectItemAtIndex(index, event);
            }
            tribute.hideMenu();
            // TODO: should fire with externalTrigger and target is outside of menu
        }
        else if (tribute.current.externalTrigger) {
            tribute.current.externalTrigger = false;
        }
        else if (tribute.current.element && !tribute.current.externalTrigger) {
            setTimeout(() => tribute.hideMenu());
        }
    }
    debounce(func, wait, immediate, ...args) {
        let timeout;
        return () => {
            const context = this;
            const later = () => {
                timeout = null;
                if (!immediate)
                    func.apply(context, args);
            };
            const callNow = immediate && !timeout;
            if (timeout !== null) {
                clearTimeout(timeout);
            }
            timeout = setTimeout(later, wait);
            if (callNow)
                func.apply(context, args);
        };
    }
}

// Thanks to https://github.com/jeff-collins/ment.io
class TributeRange {
    constructor(tribute) {
        this.tribute = tribute;
    }
    getDocument() {
        let iframe;
        if (this.tribute.current.collection) {
            iframe = this.tribute.current.collection.iframe;
        }
        if (typeof iframe === 'undefined' || iframe === null || iframe.contentWindow === null) {
            return document;
        }
        return iframe.contentWindow.document;
    }
    positionMenuAtCaret(scrollTo) {
        const context = this.tribute.current;
        const info = this.getTriggerInfo(false, this.tribute.hasTrailingSpace, true, this.tribute.allowSpaces, this.tribute.autocompleteMode);
        if (typeof (context === null || context === void 0 ? void 0 : context.element) === 'undefined' || typeof info === 'undefined')
            return;
        const coordinates = isNotContentEditable(context.element)
            ? this.getTextAreaOrInputUnderlinePosition(context.element, info.mentionPosition)
            : this.getContentEditableCaretPosition(info.mentionPosition);
        if (coordinates) {
            this.tribute.menu.positionAtCaret(info, coordinates);
        }
        if (scrollTo) {
            this.scrollIntoView();
        }
    }
    get menuContainerIsBody() {
        return this.tribute.menuContainer === document.body || !this.tribute.menuContainer;
    }
    replaceTriggerText(text, requireLeadingSpace, hasTrailingSpace, originalEvent, item) {
        var _a, _b, _c, _d;
        const info = this.getTriggerInfo(true, hasTrailingSpace, requireLeadingSpace, this.tribute.allowSpaces, this.tribute.autocompleteMode);
        const context = this.tribute.current;
        if (typeof (context === null || context === void 0 ? void 0 : context.element) === 'undefined' || typeof info === 'undefined')
            return;
        const replaceEvent = new CustomEvent('tribute-replaced', {
            detail: {
                item: item,
                instance: context,
                context: info,
                event: originalEvent,
            },
        });
        if (isNotContentEditable(context.element)) {
            const myField = context.element;
            const textSuffix = typeof this.tribute.replaceTextSuffix === 'string' ? this.tribute.replaceTextSuffix : ' ';
            const _text = text + textSuffix;
            const startPos = info.mentionPosition;
            let endPos = info.mentionPosition + (((_a = info.mentionText) === null || _a === void 0 ? void 0 : _a.length) || 0) + (textSuffix === '' ? 1 : textSuffix.length);
            if (!this.tribute.autocompleteMode) {
                endPos += (((_b = info.mentionTriggerChar) === null || _b === void 0 ? void 0 : _b.length) || 0) - 1;
            }
            myField.selectionStart = startPos + _text.length;
            myField.value = myField.value.substring(0, startPos) + _text + myField.value.substring(endPos, myField.value.length);
            myField.selectionEnd = startPos + _text.length;
        }
        else {
            let _text = text;
            if (_text instanceof HTMLElement) ;
            else {
                // add a space to the end of the pasted text
                const textSuffix = typeof this.tribute.replaceTextSuffix === 'string' ? this.tribute.replaceTextSuffix : '\xA0';
                _text += textSuffix;
            }
            let endPos = info.mentionPosition + (((_c = info.mentionText) === null || _c === void 0 ? void 0 : _c.length) || 0);
            if (!this.tribute.autocompleteMode) {
                endPos += ((_d = info.mentionTriggerChar) === null || _d === void 0 ? void 0 : _d.length) || 0;
            }
            this.pasteHtml(_text, info.mentionPosition, endPos);
        }
        context.element.dispatchEvent(new CustomEvent('input', { bubbles: true }));
        context.element.dispatchEvent(replaceEvent);
    }
    pasteHtml(htmlOrElem, startPos, endPos) {
        const sel = this.getWindowSelection();
        const range = this.getDocument().createRange();
        if (sel === null || sel.anchorNode === null)
            return;
        range.setStart(sel.anchorNode, startPos);
        range.setEnd(sel.anchorNode, endPos);
        range.deleteContents();
        const el = this.getDocument().createElement('div');
        if (htmlOrElem instanceof HTMLElement) {
            el.appendChild(htmlOrElem);
        }
        else {
            el.innerHTML = htmlOrElem;
        }
        const frag = this.getDocument().createDocumentFragment();
        let node;
        let lastNode;
        while (el.firstChild) {
            node = el.firstChild;
            lastNode = frag.appendChild(node);
        }
        range.insertNode(frag);
        // Preserve the selection
        if (lastNode) {
            const _range = range.cloneRange();
            _range.setStartAfter(lastNode);
            _range.collapse(true);
            sel === null || sel === void 0 ? void 0 : sel.removeAllRanges();
            sel === null || sel === void 0 ? void 0 : sel.addRange(_range);
        }
    }
    getWindowSelection() {
        var _a, _b;
        if ((_a = this.tribute.collection[0]) === null || _a === void 0 ? void 0 : _a.iframe) {
            const iframe = this.tribute.collection[0].iframe.contentWindow;
            if (iframe) {
                return iframe.getSelection();
            }
        }
        if ((_b = this.tribute.collection[0]) === null || _b === void 0 ? void 0 : _b.shadowRoot) {
            return this.tribute.collection[0].shadowRoot.getSelection();
        }
        return window.getSelection();
    }
    getNodePositionInParent(element) {
        if (element.parentNode === null) {
            return 0;
        }
        for (let i = 0; i < element.parentNode.childNodes.length; i++) {
            const node = element.parentNode.childNodes[i];
            if (node === element) {
                return i;
            }
        }
        return undefined;
    }
    getContentEditableSelectedPath(context) {
        const sel = this.getWindowSelection();
        if (sel === null)
            return undefined;
        let selected = sel === null || sel === void 0 ? void 0 : sel.anchorNode;
        const path = [];
        let offset;
        if (selected instanceof Node) {
            let i;
            let ce = selected instanceof HTMLElement ? selected.contentEditable : false;
            while (selected !== null && ce !== 'true') {
                i = this.getNodePositionInParent(selected);
                path.push(i);
                selected = selected.parentNode;
                if (selected instanceof HTMLElement) {
                    ce = selected.contentEditable;
                }
            }
            path.reverse();
            // getRangeAt may not exist, need alternative
            offset = sel.getRangeAt(0).startOffset;
            if (selected) {
                return {
                    selected: selected,
                    path: path,
                    offset: offset,
                };
            }
        }
        return undefined;
    }
    getTextPrecedingCurrentSelection() {
        const context = this.tribute.current;
        let text = '';
        if (typeof context.element === 'undefined')
            return text;
        if (isNotContentEditable(context.element)) {
            const textComponent = context.element;
            const startPos = textComponent.selectionStart;
            if (textComponent.value && startPos !== null && startPos >= 0) {
                text = textComponent.value.substring(0, startPos);
            }
        }
        else {
            const node = this.getWindowSelection();
            if (node === null)
                return text;
            const selectedElem = node.anchorNode;
            if (selectedElem != null) {
                const workingNodeContent = selectedElem.textContent;
                const sel = this.getWindowSelection();
                if (sel === null)
                    return;
                const selectStartOffset = sel.getRangeAt(0).startOffset;
                if (workingNodeContent && selectStartOffset >= 0) {
                    text = workingNodeContent.substring(0, selectStartOffset);
                }
            }
        }
        return text;
    }
    getLastWordInText(text) {
        let wordsArray;
        if (this.tribute.autocompleteMode) {
            if (this.tribute.autocompleteSeparator) {
                wordsArray = text.split(this.tribute.autocompleteSeparator);
            }
            else {
                wordsArray = [text];
            }
        }
        else {
            wordsArray = text.split(/\s+/);
        }
        const wordsCount = wordsArray.length - 1;
        return wordsArray[wordsCount];
    }
    getTriggerInfo(menuAlreadyActive, hasTrailingSpace, requireLeadingSpace, allowSpaces, isAutocomplete) {
        const context = this.tribute.current;
        const selectionInfo = this._getSelectionInfo(context);
        const effectiveRange = this.getTextPrecedingCurrentSelection();
        if (isAutocomplete && typeof selectionInfo !== 'undefined' && typeof effectiveRange !== 'undefined') {
            return this._getTriggerInfoWithAutocomplete(selectionInfo, effectiveRange);
        }
        if (effectiveRange === undefined || effectiveRange === null)
            return;
        const trigger = this._getTrigger(requireLeadingSpace, effectiveRange);
        if (trigger) {
            return this._getTriggerInfoNonAutocomplete(menuAlreadyActive, hasTrailingSpace, allowSpaces, effectiveRange, trigger, selectionInfo);
        }
        return undefined;
    }
    _getSelectionInfo(context) {
        if (typeof context.element === 'undefined')
            return undefined;
        if (isContentEditable(context.element)) {
            const selectionInfo = this.getContentEditableSelectedPath(context);
            if (selectionInfo) {
                return {
                    selected: selectionInfo.selected,
                    path: selectionInfo.path,
                    offset: selectionInfo.offset,
                };
            }
        }
        else {
            return { selected: context.element };
        }
        return undefined;
    }
    _getTriggerInfoWithAutocomplete(selectionInfo, effectiveRange) {
        const lastWordOfEffectiveRange = this.getLastWordInText(effectiveRange);
        return {
            mentionPosition: effectiveRange.length - (lastWordOfEffectiveRange || '').length,
            mentionText: lastWordOfEffectiveRange,
            mentionSelectedElement: selectionInfo.selected,
            mentionSelectedPath: selectionInfo.path,
            mentionSelectedOffset: selectionInfo.offset,
        };
    }
    _getTrigger(requireLeadingSpace, effectiveRange) {
        let mostRecentTriggerCharPos = -1;
        let triggerChar;
        let _requireLeadingSpace = requireLeadingSpace;
        for (const config of this.tribute.collection) {
            const c = config.trigger;
            const idx = config.requireLeadingSpace ? this.lastIndexWithLeadingSpace(effectiveRange, c) : effectiveRange.lastIndexOf(c);
            if (idx > mostRecentTriggerCharPos) {
                mostRecentTriggerCharPos = idx;
                triggerChar = c;
                _requireLeadingSpace = config.requireLeadingSpace;
            }
        }
        if (typeof triggerChar !== 'undefined' &&
            mostRecentTriggerCharPos >= 0 &&
            (mostRecentTriggerCharPos === 0 || !_requireLeadingSpace || /\s/.test(effectiveRange.substring(mostRecentTriggerCharPos - 1, mostRecentTriggerCharPos)))) {
            return {
                mostRecentTriggerCharPos,
                triggerChar,
                requireLeadingSpace: !!_requireLeadingSpace,
            };
        }
        return undefined;
    }
    _getTriggerInfoNonAutocomplete(menuAlreadyActive, hasTrailingSpace, allowSpaces, effectiveRange, trigger, selectionInfo) {
        let currentTriggerSnippet = effectiveRange.substring(trigger.mostRecentTriggerCharPos + trigger.triggerChar.length, effectiveRange.length);
        trigger.triggerChar = effectiveRange.substring(trigger.mostRecentTriggerCharPos, trigger.mostRecentTriggerCharPos + trigger.triggerChar.length);
        const firstSnippetChar = currentTriggerSnippet.substring(0, 1);
        const leadingSpace = currentTriggerSnippet.length > 0 && (firstSnippetChar === ' ' || firstSnippetChar === '\xA0');
        if (hasTrailingSpace) {
            currentTriggerSnippet = currentTriggerSnippet.trim();
        }
        const regex = allowSpaces ? /[^\S ]/g : /[\xA0\s]/g;
        this.tribute.hasTrailingSpace = regex.test(currentTriggerSnippet);
        if (!leadingSpace && (menuAlreadyActive || !regex.test(currentTriggerSnippet))) {
            return {
                mentionPosition: trigger.mostRecentTriggerCharPos,
                mentionText: currentTriggerSnippet,
                mentionSelectedElement: selectionInfo === null || selectionInfo === void 0 ? void 0 : selectionInfo.selected,
                mentionSelectedPath: selectionInfo === null || selectionInfo === void 0 ? void 0 : selectionInfo.path,
                mentionSelectedOffset: selectionInfo === null || selectionInfo === void 0 ? void 0 : selectionInfo.offset,
                mentionTriggerChar: trigger.triggerChar,
            };
        }
        return undefined;
    }
    lastIndexWithLeadingSpace(str, trigger) {
        const reversedStr = str.split('').reverse().join('');
        let index = -1;
        for (let cidx = 0, len = str.length; cidx < len; cidx++) {
            const firstChar = cidx === str.length - 1;
            const rev = reversedStr[cidx + 1];
            const leadingSpace = typeof rev === 'undefined' ? false : /\s/.test(rev);
            let match = true;
            for (let triggerIdx = trigger.length - 1; triggerIdx >= 0; triggerIdx--) {
                if (trigger[triggerIdx] !== reversedStr[cidx - triggerIdx]) {
                    match = false;
                    break;
                }
            }
            if (match && (firstChar || leadingSpace)) {
                index = str.length - 1 - cidx;
                break;
            }
        }
        return index;
    }
    isMenuOffScreen(coordinates, menuDimensions) {
        const windowWidth = window.innerWidth;
        const windowHeight = window.innerHeight;
        const doc = document.documentElement;
        const windowLeft = (window.pageXOffset || doc.scrollLeft) - (doc.clientLeft || 0);
        const windowTop = (window.pageYOffset || doc.scrollTop) - (doc.clientTop || 0);
        const menuTop = typeof coordinates.top === 'number'
            ? coordinates.top
            : typeof coordinates.bottom === 'number'
                ? windowTop + windowHeight - coordinates.bottom - menuDimensions.height
                : undefined;
        const menuRight = typeof coordinates.right === 'number' ? coordinates.right : typeof coordinates.left === 'number' ? coordinates.left + menuDimensions.width : undefined;
        const menuBottom = typeof coordinates.bottom === 'number' ? coordinates.bottom : typeof coordinates.top === 'number' ? coordinates.top + menuDimensions.height : undefined;
        const menuLeft = typeof coordinates.left === 'number'
            ? coordinates.left
            : typeof coordinates.right === 'number'
                ? windowLeft + windowWidth - coordinates.right - menuDimensions.width
                : undefined;
        return {
            top: typeof menuTop === 'number' ? menuTop < Math.floor(windowTop) : undefined,
            right: typeof menuRight === 'number' ? menuRight > Math.ceil(windowLeft + windowWidth) : undefined,
            bottom: typeof menuBottom === 'number' ? menuBottom > Math.ceil(windowTop + windowHeight) : undefined,
            left: typeof menuLeft === 'number' ? menuLeft < Math.floor(windowLeft) : undefined,
        };
    }
    getTextAreaOrInputUnderlinePosition(element, position, flipped) {
        const properties = [
            'direction',
            'boxSizing',
            'width',
            'height',
            'overflowX',
            'overflowY',
            'borderTopWidth',
            'borderRightWidth',
            'borderBottomWidth',
            'borderLeftWidth',
            'borderStyle',
            'paddingTop',
            'paddingRight',
            'paddingBottom',
            'paddingLeft',
            'fontStyle',
            'fontVariant',
            'fontWeight',
            'fontStretch',
            'fontSize',
            'fontSizeAdjust',
            'lineHeight',
            'fontFamily',
            'textAlign',
            'textTransform',
            'textIndent',
            'textDecoration',
            'letterSpacing',
            'wordSpacing',
        ];
        const div = this.getDocument().createElement('div');
        div.id = 'input-textarea-caret-position-mirror-div';
        this.getDocument().body.appendChild(div);
        const computed = getComputedStyle(element);
        div.style.whiteSpace = 'pre-wrap';
        if (element.nodeName !== 'INPUT') {
            div.style.wordWrap = 'break-word';
        }
        div.style.position = 'absolute';
        div.style.visibility = 'hidden';
        // transfer the element's properties to the div
        for (const prop of properties) {
            const value = computed.getPropertyValue(prop);
            div.style.setProperty(prop, value);
        }
        //NOT SURE WHY THIS IS HERE AND IT DOESNT SEEM HELPFUL
        // if (isFirefox) {
        //     style.width = `${(parseInt(computed.width) - 2)}px`
        //     if (element.scrollHeight > parseInt(computed.height))
        //         style.overflowY = 'scroll'
        // } else {
        //     style.overflow = 'hidden'
        // }
        const span0 = document.createElement('span');
        span0.textContent = element.value.substring(0, position);
        div.appendChild(span0);
        if (element.nodeName === 'INPUT' && div.textContent !== null) {
            div.textContent = div.textContent.replace(/\s/g, ' ');
        }
        //Create a span in the div that represents where the cursor
        //should be
        const span = this.getDocument().createElement('span');
        //we give it no content as this represents the cursor
        span.textContent = '&#x200B;';
        div.appendChild(span);
        const span2 = this.getDocument().createElement('span');
        span2.textContent = element.value.substring(position);
        div.appendChild(span2);
        const rect = element.getBoundingClientRect();
        //position the div exactly over the element
        //so we can get the bounding client rect for the span and
        //it should represent exactly where the cursor is
        div.style.position = 'fixed';
        div.style.left = `${rect.left}px`;
        div.style.top = `${rect.top}px`;
        div.style.width = `${rect.width}px`;
        div.style.height = `${rect.height}px`;
        div.scrollTop = element.scrollTop;
        const spanRect = span.getBoundingClientRect();
        this.getDocument().body.removeChild(div);
        return this.getFixedCoordinatesRelativeToRect(spanRect);
    }
    getContentEditableCaretPosition(selectedNodePosition) {
        const sel = this.getWindowSelection();
        if (sel === null || sel.anchorNode === null)
            return;
        const range = this.getDocument().createRange();
        range.setStart(sel.anchorNode, selectedNodePosition);
        range.setEnd(sel.anchorNode, selectedNodePosition);
        range.collapse(false);
        const rect = range.getBoundingClientRect();
        return this.getFixedCoordinatesRelativeToRect(rect);
    }
    getFixedCoordinatesRelativeToRect(rect) {
        const coordinates = {
            position: 'fixed',
            left: rect.left,
            top: rect.top + rect.height,
        };
        const menuDimensions = this.tribute.menu.getDimensions();
        const availableSpaceOnTop = rect.top;
        const availableSpaceOnBottom = window.innerHeight - (rect.top + rect.height);
        //check to see where's the right place to put the menu vertically
        const height = menuDimensions.height;
        if (height !== null && availableSpaceOnBottom < height) {
            if (availableSpaceOnTop >= height || availableSpaceOnTop > availableSpaceOnBottom) {
                coordinates.top = 'auto';
                coordinates.bottom = window.innerHeight - rect.top;
                if (availableSpaceOnBottom < height) {
                    coordinates.maxHeight = availableSpaceOnTop;
                }
            }
            else {
                if (availableSpaceOnTop < height) {
                    coordinates.maxHeight = availableSpaceOnBottom;
                }
            }
        }
        const availableSpaceOnLeft = rect.left;
        const availableSpaceOnRight = window.innerWidth - rect.left;
        //check to see where's the right place to put the menu horizontally
        const width = menuDimensions.width;
        if (width !== null && availableSpaceOnRight < width) {
            if (availableSpaceOnLeft >= width || availableSpaceOnLeft > availableSpaceOnRight) {
                coordinates.left = 'auto';
                coordinates.right = window.innerWidth - rect.left;
                if (availableSpaceOnRight < width) {
                    coordinates.maxWidth = availableSpaceOnLeft;
                }
            }
            else {
                if (availableSpaceOnLeft < width) {
                    coordinates.maxWidth = availableSpaceOnRight;
                }
            }
        }
        return coordinates;
    }
    scrollIntoView(elem) {
        const reasonableBuffer = 20;
        let clientRect;
        const maxScrollDisplacement = 100;
        let e = this.menu;
        if (typeof e === 'undefined')
            return;
        while (clientRect === undefined || clientRect.height === 0) {
            if (e instanceof HTMLElement) {
                clientRect = e.getBoundingClientRect();
            }
            if ((clientRect === null || clientRect === void 0 ? void 0 : clientRect.height) === 0) {
                e = e.childNodes[0];
                if (e === undefined || !('getBoundingClientRect' in e)) {
                    return;
                }
            }
        }
        const elemTop = clientRect.top;
        const elemBottom = elemTop + clientRect.height;
        if (elemTop < 0) {
            window.scrollTo(0, window.pageYOffset + clientRect.top - reasonableBuffer);
        }
        else if (elemBottom > window.innerHeight) {
            let maxY = window.pageYOffset + clientRect.top - reasonableBuffer;
            if (maxY - window.pageYOffset > maxScrollDisplacement) {
                maxY = window.pageYOffset + maxScrollDisplacement;
            }
            let targetY = window.pageYOffset - (window.innerHeight - elemBottom);
            if (targetY > maxY) {
                targetY = maxY;
            }
            window.scrollTo(0, targetY);
        }
    }
}

class TributeSearch {
    constructor(tribute) {
        this.tribute = tribute;
    }
    simpleFilter(pattern, array) {
        return array.filter((string) => {
            return this.test(pattern, string);
        });
    }
    test(pattern, string) {
        return this.match(pattern, string) !== null;
    }
    match(pattern, string, opts) {
        const _opts = opts || {};
        const pre = _opts.pre || '';
        const post = _opts.post || '';
        const compareString = (_opts.caseSensitive && string) || string.toLowerCase();
        if (_opts.skip) {
            return { rendered: string, score: 0 };
        }
        const _pattern = (_opts.caseSensitive && pattern) || pattern.toLowerCase();
        const patternCache = this.traverse(compareString, _pattern, 0, 0, []);
        if (!patternCache) {
            return null;
        }
        return {
            rendered: this.render(string, patternCache.cache, pre, post),
            score: patternCache.score,
        };
    }
    traverse(string, pattern, stringIndex, patternIndex, patternCache) {
        // if the pattern search at end
        const _pattern = this.tribute.autocompleteSeparator ? pattern.split(this.tribute.autocompleteSeparator).splice(-1)[0] : pattern;
        if (typeof _pattern === 'undefined')
            return;
        if (_pattern.length === patternIndex) {
            // calculate score and copy the cache containing the indices where it's found
            return {
                score: this.calculateScore(patternCache),
                cache: patternCache.slice(),
            };
        }
        // if string at end or remaining pattern > remaining string
        if (string.length === stringIndex || _pattern.length - patternIndex > string.length - stringIndex) {
            return undefined;
        }
        const c = _pattern[patternIndex];
        if (!c)
            return;
        let index = string.indexOf(c, stringIndex);
        let best;
        let temp;
        while (index > -1) {
            patternCache.push(index);
            temp = this.traverse(string, pattern, index + 1, patternIndex + 1, patternCache);
            patternCache.pop();
            // if downstream traversal failed, return best answer so far
            if (!temp) {
                return best;
            }
            if (!best || best.score < temp.score) {
                best = temp;
            }
            index = string.indexOf(c, index + 1);
        }
        return best;
    }
    calculateScore(patternCache) {
        let score = 0;
        let temp = 1;
        patternCache.forEach((index, i) => {
            if (i > 0) {
                const prev = patternCache[i - 1];
                if (typeof prev !== 'undefined' && prev + 1 === index) {
                    temp += 1;
                }
                else {
                    temp = 1;
                }
            }
            score += temp;
        });
        return score;
    }
    render(string, indices, pre, post) {
        let rendered = string.substring(0, indices[0]);
        indices.forEach((index, i) => {
            rendered += pre + string[index] + post + string.substring(index + 1, indices[i + 1] ? indices[i + 1] : string.length);
        });
        return rendered;
    }
    filter(pattern, arr, opts) {
        const _opts = opts || {};
        return arr
            .reduce((prev, element, idx, arr) => {
            let str = element;
            if (_opts.extract) {
                str = _opts.extract(element);
                if (!str) {
                    // take care of undefineds / nulls / etc.
                    str = '';
                }
            }
            const rendered = typeof str === 'string' ? this.match(pattern, str, _opts) : null;
            if (rendered != null) {
                prev[prev.length] = {
                    string: rendered.rendered,
                    score: rendered.score,
                    index: idx,
                    original: element,
                };
            }
            return prev;
        }, [])
            .sort((a, b) => {
            const compare = b.score - a.score;
            if (compare)
                return compare;
            return a.index - b.index;
        });
    }
}

class Tribute {
    constructor({ values = null, loadingItemTemplate = null, iframe = null, shadowRoot = null, selectClass = 'highlight', containerClass = 'tribute-container', itemClass = '', trigger = '@', autocompleteMode = false, autocompleteSeparator = /\s+/, selectTemplate = null, menuItemTemplate = null, lookup = 'key', fillAttr = 'value', collection = null, menuContainer = null, noMatchTemplate = null, requireLeadingSpace = true, allowSpaces = false, replaceTextSuffix = null, positionMenu = true, spaceSelectsMatch = false, searchOpts = {}, menuItemLimit = null, menuShowMinLength = 0, closeOnScroll = false, maxDisplayItems = null, isBlocked = false, }) {
        this._isActive = false;
        this.autocompleteMode = autocompleteMode;
        this.autocompleteSeparator = autocompleteSeparator;
        this.current = new TributeContext(this);
        this.isActive = false;
        this.menuContainer = menuContainer;
        this.allowSpaces = allowSpaces;
        this.replaceTextSuffix = replaceTextSuffix;
        this.positionMenu = positionMenu;
        this.hasTrailingSpace = false;
        this.spaceSelectsMatch = spaceSelectsMatch;
        this.closeOnScroll = closeOnScroll;
        this.menu = new TributeMenu(this);
        if (this.autocompleteMode) {
            trigger = '';
            allowSpaces = false;
        }
        if (values) {
            this.collection = [
                {
                    // symbol that starts the lookup
                    trigger: trigger,
                    // is it wrapped in an iframe
                    iframe: iframe,
                    // is it wrapped in a web component
                    shadowRoot: shadowRoot,
                    // class applied to selected item
                    selectClass: selectClass,
                    // class applied to the Container
                    containerClass: containerClass,
                    // class applied to each item
                    itemClass: itemClass,
                    // function called on select that retuns the content to insert
                    selectTemplate: selectTemplate ? selectTemplate.bind(this) : (item) => defaultSelectTemplate(this.current, item),
                    // function called that returns content for an item
                    menuItemTemplate: (menuItemTemplate || defaultMenuItemTemplate).bind(this),
                    // function called when menu is empty, disables hiding of menu.
                    noMatchTemplate: ((t) => {
                        if (typeof t === 'string') {
                            if (t.trim() === '')
                                return null;
                            return t;
                        }
                        if (typeof t === 'function') {
                            return t.bind(this);
                        }
                        return noMatchTemplate || (() => '<li>No Match Found!</li>').bind(this);
                    })(noMatchTemplate),
                    // column to search against in the object
                    lookup: lookup,
                    // column that contains the content to insert by default
                    fillAttr: fillAttr,
                    // array of objects or a function returning an array of objects
                    values: values,
                    // useful for when values is an async function
                    loadingItemTemplate: loadingItemTemplate,
                    requireLeadingSpace: requireLeadingSpace,
                    searchOpts: searchOpts,
                    menuItemLimit: menuItemLimit,
                    menuShowMinLength: menuShowMinLength,
                    // Fix for maximum number of items added to the input for the specific Collection
                    maxDisplayItems: maxDisplayItems,
                    isBlocked: isBlocked,
                },
            ];
        }
        else if (collection) {
            if (this.autocompleteMode)
                console.warn('Tribute in autocomplete mode does not work for collections');
            this.collection = collection.map((item) => {
                return {
                    trigger: item.trigger || trigger,
                    iframe: item.iframe || iframe,
                    selectClass: item.selectClass || selectClass,
                    containerClass: item.containerClass || containerClass,
                    itemClass: item.itemClass || itemClass,
                    selectTemplate: item.selectTemplate ? item.selectTemplate.bind(this) : (item) => defaultSelectTemplate(this.current, item),
                    menuItemTemplate: (item.menuItemTemplate || defaultMenuItemTemplate).bind(this),
                    // function called when menu is empty, disables hiding of menu.
                    noMatchTemplate: ((t) => {
                        if (typeof t === 'string') {
                            if (t.trim() === '')
                                return null;
                            return t;
                        }
                        if (typeof t === 'function') {
                            return t.bind(this);
                        }
                        return noMatchTemplate !== null && noMatchTemplate !== void 0 ? noMatchTemplate : (() => '<li>No Match Found!</li>').bind(this);
                    })(item.noMatchTemplate),
                    lookup: item.lookup || lookup,
                    fillAttr: item.fillAttr || fillAttr,
                    values: item.values,
                    loadingItemTemplate: item.loadingItemTemplate,
                    requireLeadingSpace: item.requireLeadingSpace,
                    searchOpts: item.searchOpts || searchOpts,
                    menuItemLimit: item.menuItemLimit || menuItemLimit,
                    menuShowMinLength: item.menuShowMinLength || menuShowMinLength,
                    // Set maximum number of items added to the input for the specific Collection
                    maxDisplayItems: item.maxDisplayItems || maxDisplayItems,
                    isBlocked: item.isBlocked || isBlocked,
                };
            });
        }
        else {
            throw new Error('[Tribute] No collection specified.');
        }
        this.range = new TributeRange(this);
        this.events = new TributeEvents(this);
        this.menuEvents = new TributeMenuEvents(this);
        this.search = new TributeSearch(this);
    }
    get isActive() {
        return this._isActive;
    }
    set isActive(val) {
        if (this._isActive !== val) {
            this._isActive = val;
            if (this.current.element) {
                const noMatchEvent = new CustomEvent(`tribute-active-${val}`);
                this.current.element.dispatchEvent(noMatchEvent);
            }
        }
    }
    triggers() {
        return this.collection.map((config) => {
            return config.trigger;
        });
    }
    attach(el) {
        // Check if it is a jQuery collection
        let _el;
        if (isJQuery(el)) {
            _el = el.get();
        }
        else {
            _el = el;
        }
        if (!_el) {
            throw new Error('[Tribute] Must pass in a DOM node or NodeList.');
        }
        // Is el an Array/Array-like object?
        if (_el instanceof NodeList || _el instanceof HTMLCollection || Array.isArray(_el)) {
            for (const e of _el) {
                if (e instanceof HTMLElement) {
                    this._attach(e);
                }
            }
        }
        else {
            this._attach(_el);
        }
    }
    _attach(el) {
        if (el.hasAttribute('data-tribute')) {
            console.warn(`Tribute was already bound to ${el.nodeName}`);
        }
        this.ensureEditable(el);
        this.events.bind(el);
        el.setAttribute('data-tribute', 'true');
    }
    ensureEditable(element) {
        if (isContentEditable(element)) {
            if (typeof element.contentEditable === 'string') {
                element.contentEditable = 'true';
            }
            else {
                throw new Error(`[Tribute] Cannot bind to ${element.nodeName}, not contentEditable`);
            }
        }
    }
    showMenuFor(element, scrollTo) {
        if (typeof this.current.collection === 'undefined')
            throw new Error('this.current.collection is undefined');
        // Check for maximum number of items added to the input for the specific Collection
        if (isMaximumItemsAdded(this.current.collection, element)) {
            //console.log("Tribute: Maximum number of items added!");
            return;
        }
        this.currentMentionTextSnapshot = this.current.mentionText;
        // create the menu if it doesn't exist.
        if (!this.menu.element) {
            const menu = this.menu.create(this.range.getDocument(), this.current.collection.containerClass);
            element.tributeMenu = menu;
            this.menuEvents.bind(menu);
        }
        this.isActive = true;
        this.menu.activate();
        this.current.process(scrollTo);
    }
    showMenuForCollection(element, collectionIndex) {
        // Check for maximum number of items added to the input for the specific Collection
        const index = collectionIndex || 0;
        const collection = this.collection[index];
        if (typeof collection === 'undefined' || isMaximumItemsAdded(collection, element)) {
            //console.log("Tribute: Maximum number of items added!");
            return;
        }
        if (element !== document.activeElement) {
            this.placeCaretAtEnd(element);
        }
        this.current.collection = collection;
        this.current.externalTrigger = true;
        this.current.element = element;
        if (element.isContentEditable) {
            this.insertTextAtCursor(this.current.collection.trigger);
        }
        else if (isNotContentEditable(element)) {
            this.insertAtCaret(element, this.current.collection.trigger);
        }
        this.showMenuFor(element);
    }
    // TODO: make sure this works for inputs/textareas
    placeCaretAtEnd(el) {
        el.focus();
        const range = document.createRange();
        range.selectNodeContents(el);
        range.collapse(false);
        const sel = window.getSelection();
        sel === null || sel === void 0 ? void 0 : sel.removeAllRanges();
        sel === null || sel === void 0 ? void 0 : sel.addRange(range);
    }
    // for contenteditable
    insertTextAtCursor(text) {
        const sel = window.getSelection();
        const range = sel === null || sel === void 0 ? void 0 : sel.getRangeAt(0);
        if (!sel || !range)
            return;
        range.deleteContents();
        const textNode = document.createTextNode(text);
        range.insertNode(textNode);
        range.selectNodeContents(textNode);
        range.collapse(false);
        sel.removeAllRanges();
        sel.addRange(range);
    }
    // for regular inputs
    insertAtCaret(textarea, text) {
        const scrollPos = textarea.scrollTop;
        let caretPos = textarea.selectionStart;
        if (!caretPos || !textarea.selectionEnd)
            return;
        const front = textarea.value.substring(0, caretPos);
        const back = textarea.value.substring(textarea.selectionEnd, textarea.value.length);
        textarea.value = front + text + back;
        caretPos = caretPos + text.length;
        textarea.selectionStart = caretPos;
        textarea.selectionEnd = caretPos;
        textarea.focus();
        textarea.scrollTop = scrollPos;
    }
    hideMenu() {
        if (this.menu.isActive) {
            this.isActive = false;
            this.menu.deactivate();
            this.current = new TributeContext(this);
        }
    }
    selectItemAtIndex(index, originalEvent) {
        const _index = Number.parseInt(index);
        if (typeof _index !== 'number' || Number.isNaN(_index) || !this.current.filteredItems || !this.current.collection || !this.current.element)
            return;
        const item = this.current.filteredItems[_index];
        const content = this.current.collection.selectTemplate(item);
        if (_index === -1 || !item) {
            const selectedNoMatchEvent = new CustomEvent('tribute-selected-no-match', { detail: content });
            this.current.element.dispatchEvent(selectedNoMatchEvent);
            return;
        }
        if (content !== null) {
            this.replaceText(content, originalEvent, item);
        }
    }
    replaceText(content, originalEvent, item) {
        this.range.replaceTriggerText(content, true, true, originalEvent, item);
    }
    _append(collection, newValues, replace) {
        if (typeof collection.values === 'function') {
            throw new Error('Unable to append to values, as it is a function.');
        }
        if (replace || collection.values === null) {
            collection.values = newValues;
        }
        else {
            collection.values = collection.values.concat(newValues);
        }
    }
    append(collectionIndex, newValues, replace) {
        const index = Number.parseInt(collectionIndex);
        if (typeof index !== 'number')
            throw new Error('please provide an index for the collection to update.');
        const collection = this.collection[index];
        if (typeof collection !== 'undefined') {
            this._append(collection, newValues, !!replace);
        }
    }
    appendCurrent(newValues, replace) {
        if (this.isActive && typeof this.current.collection !== 'undefined') {
            this._append(this.current.collection, newValues, replace);
        }
        else {
            throw new Error('No active state. Please use append instead and pass an index.');
        }
    }
    detach(el) {
        if (!el) {
            throw new Error('[Tribute] Must pass in a DOM node or NodeList.');
        }
        // Check if it is a jQuery collection
        const _el = isJQuery(el) ? el.get() : el;
        // Is el an Array/Array-like object?
        if (_el instanceof NodeList || _el instanceof HTMLCollection || Array.isArray(_el)) {
            for (const e of _el) {
                if (e instanceof HTMLElement) {
                    this._detach(e);
                }
            }
        }
        else {
            this._detach(_el);
        }
    }
    _detach(el) {
        this.events.unbind(el);
        if (el.tributeMenu) {
            this.menuEvents.unbind(el.tributeMenu);
        }
        setTimeout(() => {
            el.removeAttribute('data-tribute');
            this.isActive = false;
            if (el.tributeMenu) {
                el.tributeMenu.remove();
            }
        });
    }
    static isContentEditable(element) {
        return isContentEditable(element);
    }
}
function defaultSelectTemplate(current, item) {
    if (typeof (current === null || current === void 0 ? void 0 : current.collection) === 'undefined')
        throw new Error('current Collection is undfined');
    if (typeof item === 'undefined')
        return `${current.collection.trigger}${current.mentionText}`;
    // FIXME: should not use 'as'
    const original = item.original;
    const result = original[current.collection.fillAttr];
    if (current.element && isContentEditable(current.element)) {
        return `<span class="tribute-mention">${current.collection.trigger}${result}</span>`;
    }
    return `${current.collection.trigger}${result}`;
}
function defaultMenuItemTemplate(matchItem) {
    return matchItem.string;
}
function isMaximumItemsAdded(collection, element) {
    const result = (collection.maxDisplayItems && element.querySelectorAll(`[data-tribute-trigger="${collection.trigger}"]`).length >= collection.maxDisplayItems) ||
        collection.isBlocked;
    return !!result;
}
function isJQuery(element) {
    if (typeof element !== 'object' || element == null) {
        return false;
    }
    return 'jquery' in element && typeof element.jquery !== 'undefined';
}

/**
 * Tribute.js
 * Native ES6 JavaScript @mention Plugin
 **/

export { Tribute as default };
